from django.conf import settings
from pydjamodb.models import DynamoModel
from pynamodb.attributes import (
    MapAttribute, NumberAttribute, UnicodeAttribute, UTCDateTimeAttribute, BooleanAttribute, NumberAttribute
)


class Comment(DynamoModel):

    issue_id = UnicodeAttribute(hash_key=True)
    user_id = UnicodeAttribute(range_key=True)
    content = UnicodeAttribute()
    is_public = BooleanAttribute()
    priority = NumberAttribute()

    class Meta:
        table_name = 'comment'



from django.contrib.auth.models import User

from chamber.shortcuts import get_object_or_404

from is_core.rest.resource import CoreResource, DjangoCoreResource


class NumberOfUserIssuesResource(CoreResource,):

    def get(self):
        user = get_object_or_404(self.core.get_queryset(self.request), pk=self.kwargs.get('pk'))
        return {
            'created': user.created_issues.count(),
            'watching': user.watching_issues.count(),
        }


class UserModelResource(DjangoCoreResource):

    def watching_issues_count(self, obj):
        return obj.watching_issues.count()
    watching_issues_count.short_description = 'watching count'



import json as serializer
import os

from django.template import Library

from is_core.utils import display_object_data as utils_display_object_data


register = Library()


@register.filter
def to_list(value):
    if isinstance(value, (list, tuple)):
        return value

    return [value]


@register.filter
def json(value):
    return serializer.dumps(value)


@register.filter
def filename(value):
    return os.path.basename(value.file.name)


@register.simple_tag(takes_context=True)
def display_object_data(context, obj, field_name):
    return utils_display_object_data(obj, field_name, context.get('request'))



from elasticsearch import NotFoundError

from is_core.generic_views.detail_views import ModelReadonlyDetailView
from is_core.generic_views.inlines.inline_table_views import BaseModelInlineTableView
from is_core.generic_views.table_views import BaseModelTableView
from is_core.utils.field_api import BaseModelFieldDescriptor, pretty_name

from elasticsearch_dsl import Document


class ElasticsearcDocumenthFieldDescriptor(BaseModelFieldDescriptor):

    is_model_field = True
    base_model = Document

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        model_field = model._doc_type.mapping.properties.to_dict()['properties'].get(field_name)
        return cls(model, field_name, model_field) if model_field else None

    def get_label(self):
        return getattr(self.model_field_or_method, 'short_description', pretty_name(self.field_name))

    def get_value(self, instance, request=None):
        return getattr(instance, self.field_name)


class ElasticsearchTableViewMixin:

    list_verbose_name = '%(verbose_name_plural)s'

    @property
    def model_name(self):
        return str(self.model.__class__.__name__.lower())


class ElasticsearchInlineTableView(ElasticsearchTableViewMixin, BaseModelInlineTableView):
    pass


class ElasticsearchTableView(ElasticsearchTableViewMixin, BaseModelTableView):

    add_button_verbose_name = 'add %(verbose_name)s'


class ElasticsearchDetailView(ModelReadonlyDetailView):

    detail_verbose_name = '%(verbose_name)s'

    def _get_obj_or_none(self):
        if not self.kwargs.get('pk'):
            return None
        try:
            return self.model.get(id=self.kwargs.get('pk'))
        except NotFoundError:
            return None



"""
Core of django-is-core.
Contains controller added between model and UI/REST.
"""
import sys

from copy import deepcopy

from collections import OrderedDict

from django.utils.translation import ugettext_lazy as _
from django.utils.functional import cached_property
from django.urls import reverse

import import_string

from is_core.auth.permissions import FieldsSetPermission
from is_core.config import settings
from is_core.actions import WebAction, ConfirmRestAction
from is_core.generic_views.form_views import BulkChangeFormView
from is_core.generic_views.detail_views import DjangoDetailFormView
from is_core.generic_views.add_views import DjangoAddFormView
from is_core.generic_views.table_views import DjangoTableView
from is_core.rest.resource import DjangoCoreResource
from is_core.rest.paginators import DjangoOffsetBasedPaginator
from is_core.patterns import UiPattern, RestPattern, DoubleRestPattern
from is_core.utils import flatten_fieldsets, GetMethodFieldMixin, get_model_name, PK_PATTERN
from is_core.menu import LinkMenuItem
from is_core.loading import register_core
from is_core.rest.factory import modelrest_factory
from is_core.forms.models import SmartModelForm
from is_core.utils.decorators import short_description

from .auth.permissions import PermissionsSet, IsAdminUser


class CoreBase(type):
    """Metaclass for IS core classes. Its main purpose is automatic registration cores to your application."""

    def __new__(cls, *args, **kwargs):
        name, _, attrs = args

        abstract = attrs.pop('abstract', False)
        super_new = super(CoreBase, cls).__new__
        new_class = super_new(cls, *args, **kwargs)
        model_module = sys.modules[new_class.__module__]
        app_label = model_module.__name__.split('.')[-2]

        if name != 'NewBase' and not abstract and new_class.register:
            register_core(app_label, new_class)
        return new_class


class Core(metaclass=CoreBase):
    """
    Parent of all IS cores. Contains common methods for all cores.
    This class is abstract.
    """

    abstract = True
    register = True

    list_actions = ()

    menu_group = None

    can_create = False
    can_read = True
    can_update = False
    can_delete = False

    default_permission = IsAdminUser()

    permission = None
    field_permissions = FieldsSetPermission()

    verbose_name = None
    verbose_name_plural = None

    def __init__(self, site_name, menu_parent_groups):
        self.site_name = site_name
        self.menu_parent_groups = menu_parent_groups
        self.permission = self._init_permission(self.permission)

    def _init_permission(self, permission):
        return self._generate_permission_set() if permission is None else deepcopy(permission)

    def _get_default_permission(self, name):
        return self.default_permission

    def _generate_permission_set(self):
        permission_dict = {}
        for permission_name in ('create', 'read', 'update', 'delete'):
            if getattr(self, 'can_{}'.format(permission_name)):
                permission_dict[permission_name] = self._get_default_permission(permission_name)
        return PermissionsSet(**permission_dict)

    def init_request(self, request):
        pass

    def get_urlpatterns(self, patterns):
        urls = []
        for pattern in patterns.values():
            url = pattern.get_url()
            if url:
                urls.append(url)

        return urls

    def get_urls(self):
        return ()

    def get_menu_groups(self):
        menu_groups = list(self.menu_parent_groups)
        if self.menu_group:
            menu_groups += (self.menu_group,)
        return menu_groups

    def get_url_prefix(self):
        return '/'.join(self.get_menu_groups())

    def get_menu_group_pattern_name(self):
        return '-'.join(self.get_menu_groups())

    def get_verbose_name(self):
        return self.verbose_name

    def get_verbose_name_plural(self):
        return self.verbose_name_plural


class ModelCore(GetMethodFieldMixin, Core):

    abstract = True

    list_actions = ()

    register_model = True

    fields = ()
    list_fields = None

    default_ordering = None

    field_labels = {}

    can_read = True
    can_create = False
    can_update = False
    can_delete = False

    menu_group = None

    def get_fields(self, request, obj=None):
        return self.fields

    def get_field_labels(self, request):
        return self.field_labels

    def get_default_ordering(self):
        return self.default_ordering

    def get_queryset(self, request):
        raise NotImplementedError

    def get_list_actions(self, request, obj):
        return list(self.list_actions)

    def get_default_action(self, request, obj):
        return None


class DjangoCore(ModelCore):
    """
    Parent of REST and UI cores that works as controller to specific model.
    This class is abstract.
    """

    abstract = True

    # form params
    form_class = SmartModelForm
    form_edit_class = None
    form_add_class = None

    can_read = True
    can_create = True
    can_update = True
    can_delete = True

    def get_verbose_name(self):
        return self.model._meta.verbose_name if self.verbose_name is None else self.verbose_name

    def get_verbose_name_plural(self):
        return self.model._meta.verbose_name_plural if self.verbose_name_plural is None else self.verbose_name_plural

    def get_form_class(self, request, obj=None):
        return self.form_class

    def get_form_edit_class(self, request, obj=None):
        return self.form_edit_class or self.get_form_class(request, obj)

    def get_form_add_class(self, request, obj=None):
        return self.form_add_class or self.get_form_class(request, obj)

    def pre_save_model(self, request, obj, form, change):
        pass

    def save_model(self, request, obj, form, change):
        obj.save()

    def post_save_model(self, request, obj, form, change):
        pass

    def pre_delete_model(self, request, obj):
        pass

    def delete_model(self, request, obj):
        obj.delete()

    def post_delete_model(self, request, obj):
        pass

    @property
    def menu_group(self):
        return get_model_name(self.model)

    def get_default_ordering(self):
        return self.default_ordering if self.default_ordering is not None else (self.model._meta.ordering or ('pk',))

    def preload_queryset(self, request, qs):
        return qs

    def get_queryset(self, request):
        return self.model._default_manager.get_queryset().order_by(*self.get_default_ordering())


class UiCore(Core):
    """Main core for UI views."""

    abstract = True

    show_in_menu = True
    menu_url_name = None
    view_classes = ()
    default_ui_pattern_class = UiPattern

    def init_ui_request(self, request):
        self.init_request(request)

    def get_view_classes(self):
        return list(self.view_classes)

    @cached_property
    def ui_patterns(self):
        return self.get_ui_patterns()

    def get_ui_patterns(self):
        ui_patterns = OrderedDict()
        for view_class_definition in self.get_view_classes():
            name, view_vals = (view_class_definition[0], view_class_definition[1:])
            if name not in ui_patterns:
                if len(view_vals) == 3:
                    pattern, view, ViewPatternClass = view_vals
                else:
                    pattern, view = view_vals
                    ViewPatternClass = self.default_ui_pattern_class

                if isinstance(view, str):
                    view = getattr(self, view)()

                pattern_names = [name]
                group_pattern_name = self.get_menu_group_pattern_name()
                if group_pattern_name:
                    pattern_names += [self.get_menu_group_pattern_name()]

                ui_patterns[name] = ViewPatternClass('-'.join(pattern_names), self.site_name, pattern, view, self)
        return ui_patterns

    def get_urls(self):
        return self.get_urlpatterns(self.ui_patterns)

    def get_show_in_menu(self, request):
        return (
            self.show_in_menu and self.menu_url_name and self.menu_url_name in self.ui_patterns and
            self.ui_patterns.get(self.menu_url_name).has_permission('get', request)
        )

    def is_active_menu_item(self, request, active_group):
        return active_group == self.menu_group

    def get_menu_item(self, request, active_group, item_class=LinkMenuItem):
        if self.get_show_in_menu(request):
            return item_class(self.get_verbose_name_plural(), self.menu_url(request),
                              self.menu_group, self.is_active_menu_item(request, active_group))

    def menu_url(self, request):
        return self.ui_patterns.get(self.menu_url_name).get_url_string(request)


class RestCore(Core):
    """Main core for REST views."""

    rest_classes = ()
    default_rest_pattern_class = RestPattern
    abstract = True

    api_url_name = None

    def init_rest_request(self, request):
        self.init_request(request)

    def get_rest_classes(self):
        return list(self.rest_classes)

    @cached_property
    def rest_patterns(self):
        return self.get_rest_patterns()

    def get_rest_patterns(self):
        rest_patterns = OrderedDict()
        for rest_class_definition in self.get_rest_classes():
            name, rest_vals = (rest_class_definition[0], rest_class_definition[1:])
            if name not in rest_patterns:
                if len(rest_vals) == 3:
                    pattern, rest, RestPatternClass = rest_vals
                else:
                    pattern, rest = rest_vals
                    RestPatternClass = self.default_rest_pattern_class

                if isinstance(rest, str):
                    rest = getattr(self, rest)()

                pattern_names = [name]
                group_pattern_name = self.get_menu_group_pattern_name()
                if group_pattern_name:
                    pattern_names += [self.get_menu_group_pattern_name()]
                rest_patterns[name] = RestPatternClass('-'.join(pattern_names), self.site_name, pattern, rest, self)
        return rest_patterns

    def get_urls(self):
        return self.get_urlpatterns(self.rest_patterns)

    def get_api_url_name(self):
        return self.api_url_name or '{}:api-{}'.format(self.site_name, self.get_menu_group_pattern_name())

    def get_api_url(self, request):
        return reverse(self.get_api_url_name())

    def get_api_detail_url_name(self):
        return self.api_url_name or '{}:api-resource-{}'.format(self.site_name, self.get_menu_group_pattern_name())

    def get_api_detail_url(self, request, obj):
        return reverse(self.get_api_detail_url_name(), kwargs={'pk': obj.pk})


class UiRestCoreMixin:
    """Helper that joins urls generated wit REST core and UI core."""

    def get_urls(self):
        return self.get_urlpatterns(self.rest_patterns) + self.get_urlpatterns(self.ui_patterns)


class UiRestCore(UiRestCoreMixin, UiCore, RestCore):
    """UI REST Core, its main purpose is create custom REST resources and UI views."""

    abstract = True


class HomeUiCore(UiCore):

    menu_url_name = 'index'
    verbose_name_plural = _('Home')
    menu_group = 'home'
    abstract = settings.HOME_CORE != 'is_core.main.HomeUiCore'

    def get_view_classes(self):
        view_classes = super().get_view_classes()
        view_classes.append(('index', r'', import_string(settings.HOME_VIEW)))
        return view_classes

    def menu_url(self, request):
        return '/'

    def get_url_prefix(self):
        return ''


class ModelUiCore(ModelCore, UiCore):

    abstract = True

    menu_url_name = 'list'

    # list view params
    list_per_page = None
    default_list_filter = {}

    # add/edit view params
    fieldsets = None
    readonly_fields = None
    inline_views = None

    list_fields = None
    export_fields = None
    list_view_export_fields = None
    detail_view_export_fields = None
    export_types = settings.EXPORT_TYPES

    ui_add_view = None
    ui_detail_view = None
    ui_list_view = None

    @cached_property
    def default_model_view_classes(self):
        default_model_view_classes = []

        add_view = self.get_ui_add_view()
        detail_view = self.get_ui_detail_view()
        list_view = self.get_ui_list_view()

        if self.can_create and add_view:
            default_model_view_classes.append(('add', r'add/', add_view))
        if (self.can_read or self.can_update) and detail_view:
            default_model_view_classes.append(('detail', r'{}/'.format(PK_PATTERN), detail_view))
        if self.can_read and list_view:
            default_model_view_classes.append(('list', r'', list_view))
        return default_model_view_classes

    def get_view_classes(self):
        return super().get_view_classes() + list(self.default_model_view_classes)

    def get_ui_add_view(self):
        return self.ui_add_view

    def get_ui_detail_view(self):
        return self.ui_detail_view

    def get_ui_list_view(self):
        return self.ui_list_view

    def get_fieldsets(self, request, obj=None):
        return self.fieldsets

    def get_fields(self, request, obj=None):
        fieldsets = self.get_fieldsets(request, obj)
        return flatten_fieldsets(fieldsets) if fieldsets is not None else self.fields

    def get_readonly_fields(self, request, obj=None):
        return [] if self.readonly_fields is None else self.readonly_fields

    def get_urls(self):
        return self.get_urlpatterns(self.ui_patterns)

    def get_show_in_menu(self, request):
        return (
            self.menu_url_name in self.ui_patterns and self.show_in_menu and
            self.ui_patterns.get(self.menu_url_name).has_permission('get', request)
        )

    def get_inline_views(self, request, obj=None):
        return self.inline_views

    def get_default_list_filter(self, request):
        return self.default_list_filter.copy()

    def get_list_fields(self, request):
        return list(self.list_fields) if self.list_fields is not None else self.get_fields(request)

    def get_export_fields(self, request, obj=None):
        return list(self.export_fields) if self.export_fields is not None else self.get_list_fields(request)

    def get_export_types(self, request, obj=None):
        return list(self.export_types or ())

    def get_list_per_page(self, request):
        return self.list_per_page

    def get_add_url(self, request):
        return self.ui_patterns.get('add').get_url_string(request) if 'add' in self.ui_patterns else None

    @short_description(_('object name'))
    def _obj_name(self, obj):
        return str(obj)


class DjangoUiCore(DjangoCore, ModelUiCore):
    """Main core controller for specific model that provides UI views for model management (add, edit, list)."""

    abstract = True

    ui_add_view = DjangoAddFormView
    ui_detail_view = DjangoDetailFormView
    ui_list_view = DjangoTableView


class ModelRestCore(RestCore, ModelCore):
    """
    Main core controller for specific model that provides REST resources for model management.
    CRUD (POST, GET, PUT, DELETE).
    """

    abstract = True

    # Allowed rest fields
    rest_fields = None
    rest_extra_fields = None
    rest_detailed_fields = None
    rest_general_fields = None
    rest_guest_fields = None
    rest_default_fields = ('_obj_name',)
    rest_filter_fields = None
    rest_filter_manager = None
    rest_extra_filter_fields = None
    rest_order_fields = None
    rest_extra_order_fields = None
    rest_register = True

    rest_form_edit_class = None
    rest_form_add_class = None

    rest_resource_class = None
    rest_field_labels = None
    rest_paginator = None

    def get_rest_allowed_methods(self):
        rest_allowed_methods = ['options']
        if self.can_read:
            rest_allowed_methods += ['get', 'head']
        if self.can_update:
            rest_allowed_methods += ['put', 'patch']
        if self.can_create:
            rest_allowed_methods.append('post')
        if self.can_delete:
            rest_allowed_methods.append('delete')
        return rest_allowed_methods

    def get_rest_field_labels(self, request):
        return (
            self.rest_field_labels if self.rest_field_labels is not None
            else self.get_field_labels(request)
        )

    def get_rest_form_add_class(self, request, obj=None):
        return self.rest_form_add_class or self.get_form_add_class(request, obj)

    def get_rest_form_edit_class(self, request, obj=None):
        return self.rest_form_edit_class or self.get_form_edit_class(request, obj)

    def get_rest_form_fields(self, request, obj=None):
        return self.get_fields(request, obj)

    def get_rest_form_exclude(self, request, obj=None):
        return []

    def get_rest_fields(self, request, obj=None):
        return self.rest_fields

    def get_rest_extra_fields(self, request, obj=None):
        return list(self.rest_extra_fields or ())

    def get_rest_detailed_fields(self, request, obj=None):
        return list(self.rest_detailed_fields or ())

    def get_rest_general_fields(self, request, obj=None):
        return list(self.rest_general_fields or ())

    def get_rest_guest_fields(self, request, obj=None):
        return list(self.rest_guest_fields or ())

    def get_rest_default_fields(self, request, obj=None):
        return list(self.rest_default_fields or ())

    def get_rest_filter_fields(self, request):
        return self.rest_filter_fields

    def get_rest_extra_filter_fields(self, request):
        return self.rest_extra_filter_fields

    def get_rest_order_fields(self, request):
        return self.rest_order_fields

    def get_rest_extra_order_fields(self, request):
        return self.rest_extra_order_fields

    def get_rest_class(self):
        resource_kwargs = {}
        resource_kwargs['paginator'] = self.rest_paginator
        if self.rest_filter_manager:
            resource_kwargs['filter_manager'] = self.rest_filter_manager
        return modelrest_factory(
            self.model, self.rest_resource_class,
            resource_kwargs=resource_kwargs,
            register=self.rest_register
        )

    def get_rest_patterns(self):
        rest_patterns = super().get_rest_patterns()
        rest_patterns.update(
            DoubleRestPattern(
                self.get_rest_class(), self.default_rest_pattern_class, self,
                self.get_rest_allowed_methods()
            ).patterns
        )
        return rest_patterns

    def get_list_actions(self, request, obj):
        list_actions = super().get_list_actions(request, obj)
        api_resource = self.rest_patterns.get('api-resource')
        if self.can_delete and api_resource.has_permission('delete', request, obj=obj):
            confirm_dialog = ConfirmRestAction.ConfirmDialog(_('Do you really want to delete "%s"') % obj)
            list_actions.append(ConfirmRestAction('api-resource-{}'.format(self.get_menu_group_pattern_name()),
                                                  _('Delete'), 'DELETE', confirm_dialog=confirm_dialog,
                                                  class_name='delete', success_text=_('Record "%s" was deleted') % obj))
        return list_actions

    # Resource extra fields

    def _rest_links(self, obj, request):
        rest_links = {}
        for pattern in self.rest_patterns.values():
            if pattern.send_in_rest:
                url = pattern.get_url_string(request, obj=obj)
                if url:
                    allowed_methods = pattern.get_allowed_methods(request, obj)
                    if allowed_methods:
                        rest_links[pattern.name] = {
                            'url': url,
                            'methods': [method.upper() for method in allowed_methods]
                        }
        return rest_links

    def _actions(self, obj, request):
        return self.get_list_actions(request, obj)

    def _default_action(self, obj, request):
        return self.get_default_action(request, obj=obj)


class DjangoRestCore(ModelRestCore, DjangoCore):
    """
    Main core controller for specific model that provides REST resources for model management.
    CRUD (POST, GET, PUT, DELETE).
    """

    abstract = True

    rest_resource_class = DjangoCoreResource
    rest_paginator = DjangoOffsetBasedPaginator()

    def get_rest_general_fields(self, request, obj=None):
        return list(
            self.model._rest_meta.general_fields if self.rest_general_fields is None
            else self.rest_general_fields
        )

    def get_rest_detailed_fields(self, request, obj=None):
        return list(
            self.model._rest_meta.detailed_fields if self.rest_detailed_fields is None
            else self.rest_detailed_fields
        )

    def get_rest_guest_fields(self, request, obj=None):
        return list(
            self.model._rest_meta.guest_fields if self.rest_guest_fields is None
            else self.rest_guest_fields
        )

    def get_rest_extra_filter_fields(self, request):
        return (
            self.model._rest_meta.extra_filter_fields if self.rest_extra_filter_fields is None
            else self.rest_extra_filter_fields
        )

    def get_rest_filter_fields(self, request):
        return (
            self.model._rest_meta.filter_fields if self.rest_filter_fields is None
            else self.rest_filter_fields
        )

    def get_rest_extra_order_fields(self, request):
        return (
            self.model._rest_meta.extra_order_fields if self.rest_extra_order_fields is None
            else self.rest_extra_order_fields
        )

    def get_rest_order_fields(self, request):
        return (
            self.model._rest_meta.order_fields if self.rest_order_fields is None
            else self.rest_order_fields
        )


class ModelUiRestCore(UiRestCoreMixin, ModelRestCore, ModelUiCore):

    abstract = True

    default_rest_extra_fields = (
        '_web_links', '_rest_links', '_default_action', '_actions', '_class_names', '_obj_name'
    )

    rest_obj_class_names = ()

    def get_rest_extra_fields(self, request, obj=None):
        return (
            super().get_rest_extra_fields(request, obj) +
            list(self.get_list_fields(request)) +
            list(self.get_export_fields(request)) +
            list(self.default_rest_extra_fields)
        )

    def get_list_actions(self, request, obj):
        list_actions = super().get_list_actions(request, obj)
        detail_pattern = self.ui_patterns.get('detail')
        if detail_pattern and detail_pattern.has_permission('get', request, obj=obj):
            return [
                WebAction(
                    'detail-{}'.format(self.get_menu_group_pattern_name()), _('Detail'),
                    'edit' if detail_pattern.has_permission('post', request, obj=obj) else 'detail'
                )
            ] + list(list_actions)
        else:
            return list_actions

    def get_default_action(self, request, obj):
        return 'detail-{}'.format(self.get_menu_group_pattern_name())

    def web_link_patterns(self, request):
        return self.ui_patterns.values()

    def get_rest_obj_class_names(self, request, obj):
        return list(self.rest_obj_class_names)

    # Resource extra fields

    def _web_links(self, obj, request):
        web_links = {}
        for pattern in self.web_link_patterns(request):
            if pattern.send_in_rest:
                url = pattern.get_url_string(request, obj=obj)
                if url and pattern.has_permission('get', request, obj=obj):
                    web_links[pattern.name] = url
        return web_links

    def _class_names(self, obj, request):
        return self.get_rest_obj_class_names(request, obj)


class DjangoUiRestCore(DjangoRestCore, DjangoUiCore, ModelUiRestCore):
    """
    Combination of UI views and REST resources. UI views uses REST resources for printing model data, filtering,
    paging and so on.
    """

    abstract = True

    bulk_change_url_name = 'bulk-change'
    bulk_change_fields = ()

    bulk_change_enabled = False

    def get_rest_form_fields(self, request, obj=None):
        return list(set(super().get_rest_form_fields(request, obj)) | set(self.get_bulk_change_fields(request)))

    def get_rest_form_exclude(self, request, obj=None):
        return self.get_readonly_fields(request, obj)

    def get_view_classes(self):
        view_classes = super().get_view_classes()
        if self.is_bulk_change_enabled():
            view_classes.append((self.get_bulk_change_url_name(), r'bulk-change/?', BulkChangeFormView))
        return view_classes

    def is_bulk_change_enabled(self):
        return self.bulk_change_enabled

    def get_bulk_change_url_name(self):
        return self.bulk_change_url_name

    def get_bulk_change_fields(self, request):
        return self.bulk_change_fields



from is_core.contrib.elasticsearch.cores import ElasticsearchUiRestCore

from .models import Comment


class ElasticsearchCommentCore(ElasticsearchUiRestCore):

    model = Comment
    menu_group = 'elasticsearch-comment'

    fields = ('id', 'user_id', 'content', 'is_public', 'priority')

    verbose_name = 'comment'
    verbose_name_plural = 'comments'



# Generated by Django 2.2.4 on 2019-08-05 20:01

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('created_by',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='created_issues',
                                   to=settings.AUTH_USER_MODEL, verbose_name='Created by')),
                ('leader',
                 models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='leading_issue',
                                      to=settings.AUTH_USER_MODEL, verbose_name='Leader')),
                ('solver', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                                related_name='solving_issue', to=settings.AUTH_USER_MODEL,
                                                verbose_name='Solver')),
                ('watched_by',
                 models.ManyToManyField(blank=True, related_name='watching_issues', to=settings.AUTH_USER_MODEL,
                                        verbose_name='Watched by')),
            ],
        ),
    ]



from germanium.decorators import login
from germanium.test_cases.rest import RestTestCase
from germanium.tools import assert_equal, assert_false
from germanium.tools.http import (assert_http_forbidden, assert_http_unauthorized, assert_http_accepted,
                                  assert_http_not_found, assert_http_method_not_allowed)
from germanium.tools.rest import assert_valid_JSON_response, assert_valid_JSON_created_response

from .factories import IssueFactory, UserFactory
from .test_case import HelperTestCase, AsSuperuserTestCase

from issue_tracker.models import Issue


__all__ =(
    'RESTPermissionsTestCase',
)


class RESTPermissionsTestCase(AsSuperuserTestCase, HelperTestCase, RestTestCase):

    USER_API_URL = '/api/user/'
    ISSUE_API_URL = '/api/issue/'
    USER_ISSUES_API_URL = '/api/user/%(user_pk)s/issue-number/'

    def test_non_logged_user_should_receive_401(self):
        resp = self.get(self.USER_API_URL)
        assert_http_unauthorized(resp)

    @login(is_superuser=True)
    def test_superuser_may_read_user_data(self):
        [self.get_user_obj() for _ in range(5)]

        resp = self.get(self.USER_API_URL)
        assert_valid_JSON_response(resp)
        output = self.deserialize(resp)
        assert_equal(len(output), 6)

    @login(is_superuser=False)
    def test_user_can_read_all_user_data(self):
        [self.get_user_obj() for _ in range(5)]
        resp = self.get(f'{self.USER_API_URL}?_fields=id')
        assert_valid_JSON_response(resp)
        output = self.deserialize(resp)
        assert_equal(len(output), 1)
        assert_equal(output[0]['id'], self.logged_user.user.pk)

    @login(is_superuser=False)
    def test_user_can_read_only_itself(self):
        resp = self.get(('%s%s/') % (self.USER_API_URL, self.logged_user.user.pk))
        assert_valid_JSON_response(resp)

        user = self.get_user_obj()
        resp = self.get(('%s%s/') % (self.USER_API_URL, user.pk))
        assert_http_not_found(resp)

    @login(is_superuser=True)
    def test_superuser_can_add_new_user(self):
        resp = self.post(self.USER_API_URL, data=self.get_user_data())
        assert_valid_JSON_created_response(resp)

    @login(is_superuser=False)
    def test_only_superuser_can_add_new_user(self):
        resp = self.post(self.USER_API_URL, data=self.get_user_data())
        assert_http_forbidden(resp)

    @login(is_superuser=True)
    def test_superuser_can_add_update_user(self):
        user = self.get_user_obj()
        resp = self.put('%s%s/' % (self.USER_API_URL, user.pk), data={})
        assert_valid_JSON_response(resp)

    @login(is_superuser=False)
    def test_user_can_update_only_itself(self):
        user = self.get_user_obj()
        resp = self.put('%s%s/' % (self.USER_API_URL, user.pk), data={})
        assert_http_not_found(resp)

        user = self.logged_user.user
        resp = self.put('%s%s/' % (self.USER_API_URL, user.pk), data={})
        assert_valid_JSON_response(resp)

    @login(is_superuser=True)
    def test_superuser_can_delete_new_user(self):
        user = self.get_user_obj()
        resp = self.delete('%s%s/' % (self.USER_API_URL, user.pk))
        assert_http_accepted(resp)

    @login(is_superuser=False)
    def test_only_superuser_can_delete_new_user(self):
        user = self.get_user_obj()
        resp = self.delete('%s%s/' % (self.USER_API_URL, user.pk))
        assert_http_forbidden(resp)

    @login(is_superuser=False)
    def test_superuser_should_not_delete_another_superuser(self):
        user = self.get_user_obj(is_superuser=True)
        resp = self.delete('%s%s/' % (self.USER_API_URL, user.pk))
        assert_http_forbidden(resp)

    def test_not_logged_user_can_not_get_number_of_user_issues(self):
        user = self.get_user_obj()
        resp = self.get(self.USER_ISSUES_API_URL % {'user_pk': user.pk})
        assert_http_unauthorized(resp)

    @login(is_superuser=True)
    def test_superuser_can_get_number_of_user_issues(self):
        user = self.get_user_obj()
        resp = self.get(self.USER_ISSUES_API_URL % {'user_pk': user.pk})
        assert_valid_JSON_response(resp)

        output = self.deserialize(resp)
        assert_equal(output['created'], 0)
        assert_equal(output['watching'], 0)

    @login(is_superuser=False)
    def test_only_superuser_can_get_number_of_other_users_issues(self):
        user = self.get_user_obj()
        resp = self.get(self.USER_ISSUES_API_URL % {'user_pk': user.pk})
        assert_http_not_found(resp)

    @login(is_superuser=False)
    def test_user_can_get_number_of_its_issues(self):
        user = self.logged_user.user
        resp = self.get(self.USER_ISSUES_API_URL % {'user_pk': user.pk})
        assert_valid_JSON_response(resp)
        output = self.deserialize(resp)
        assert_equal(output['created'], 0)
        assert_equal(output['watching'], 0)

    @login(is_superuser=True)
    def test_issue_delete_should_not_be_supported_because_can_delete_is_disabled(self):
        assert_http_method_not_allowed(self.delete(self.ISSUE_API_URL + '1/'))

    @login(is_superuser=True)
    def test_issue_post_should_not_be_supported_because_can_create_is_disabled(self):
        assert_http_method_not_allowed(self.post(self.ISSUE_API_URL, data={}))

    @login(is_superuser=True)
    def test_issue_actions_should_not_contains_edit_because_update_is_disabled(self):
        IssueFactory()
        resp = self.get(self.ISSUE_API_URL+'?_fields=_actions')
        assert_equal(len(resp.json()[0]['_actions']), 1)
        assert_equal(resp.json()[0]['_actions'][0]['class_name'], 'detail')

    @login(is_superuser=True)
    def test_new_user_actions_should_contains_edit_and_delete_because_update_is_disabled(self):
        UserFactory()
        resp = self.get(self.USER_API_URL+'?_fields=_actions')
        assert_equal(len(resp.json()[1]['_actions']), 2)
        assert_equal({action['class_name'] for action in resp.json()[1]['_actions']}, {'edit', 'delete'})

    @login(is_superuser=True)
    def test_logged_user_actions_should_contains_only_edit_action(self):
        resp = self.get(self.USER_API_URL+'?_fields=_actions')
        assert_equal(len(resp.json()[0]['_actions']), 1)
        assert_equal({action['class_name'] for action in resp.json()[0]['_actions']}, {'edit'})

    @login(is_superuser=True)
    def test_superuser_should_read_username_and_is_superuser_fields(self):
        resp = self.get(self.USER_API_URL+'?_fields=username,is_superuser')
        assert_equal(set(resp.json()[0].keys()), {'username', 'is_superuser'})

    @login(is_superuser=False)
    def test_user_should_not_read_read_is_superuser_fields(self):
        resp = self.get(self.USER_API_URL+'?_fields=username,is_superuser')
        assert_equal(set(resp.json()[0].keys()), {'username'})

    @login(is_superuser=False)
    def test_user_should_not_update_is_superuser(self):
        user = self.logged_user.user
        resp = self.put('%s%s/' % (self.USER_API_URL, user.pk), data={'is_superuser': True})
        assert_valid_JSON_response(resp)
        user.refresh_from_db()
        assert_false(user.is_superuser)

    @login(is_superuser=True)
    def test_superuser_should_update_is_superuser(self):
        user = self.logged_user.user
        resp = self.put('%s%s/' % (self.USER_API_URL, user.pk), data={'is_superuser': False})
        assert_valid_JSON_response(resp)
        user.refresh_from_db()
        assert_false(user.is_superuser)



from django.db import models
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

from is_core.utils.decorators import relation


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class Issue(models.Model):

    name = models.CharField(verbose_name='Name', max_length=100, null=False, blank=False)
    watched_by = models.ManyToManyField(AUTH_USER_MODEL, verbose_name='Watched by', blank=True,
                                        related_name='watching_issues')
    created_by = models.ForeignKey(AUTH_USER_MODEL, verbose_name='Created by', null=False, blank=False,
                                   related_name='created_issues', on_delete=models.CASCADE)

    solver = models.OneToOneField(AUTH_USER_MODEL, verbose_name='Solver', null=True, blank=True,
                                  related_name='solving_issue', on_delete=models.CASCADE)
    leader = models.OneToOneField(AUTH_USER_MODEL, verbose_name='Leader', null=False, blank=False,
                                  related_name='leading_issue', on_delete=models.CASCADE)

    related_object_ct = models.ForeignKey(ContentType, null=True, blank=True, on_delete=models.CASCADE)
    related_object_id = models.PositiveIntegerField(null=True, blank=True)
    related_object = GenericForeignKey('related_object_ct', 'related_object_id')

    is_issue = True


    def watched_by_string(self):
        return ', '.join(str(user) for user in self.watched_by.all())
    watched_by_string.order_by = 'watched_by'

    @property
    @relation(AUTH_USER_MODEL)
    def watched_by_method(self):
        return list(self.watched_by.all())

    def __unicode__(self):
        return 'issue: %s' % self.name



from django.http.response import Http404
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import cached_property
from django.views.generic.base import TemplateView
from django.template import RequestContext

from chamber.shortcuts import get_object_or_none

from is_core.auth.views import FieldPermissionViewMixin
from is_core.auth.permissions import (
    PermissionsSet, CoreReadAllowed, CoreUpdateAllowed, CoreAllowed, DEFAULT_PERMISSION
)
from is_core.generic_views.base import DefaultModelCoreViewMixin
from is_core.utils import (
    get_readonly_field_data, get_export_types_with_content_type, GetMethodFieldMixin, display_for_value,
    get_fieldsets_without_disallowed_fields, get_inline_views_opts_from_fieldsets
)
from is_core.generic_views.mixins import ListParentMixin, GetDjangoObjectCoreViewMixin, GetModelObjectCoreViewMixin
from is_core.generic_views.inlines.inline_table_views import DjangoInlineTableView
from is_core.rest.datastructures import ModelFlatRestFields

from .form_views import DjangoCoreFormView


class DjangoDetailFormView(GetDjangoObjectCoreViewMixin, DjangoCoreFormView):

    template_name = 'is_core/generic_views/detail_form.html'
    form_template = 'is_core/forms/model_detail_form.html'
    view_type = 'detail'
    messages = {'success': _('The %(name)s "%(obj)s" was changed successfully.'),
                'error': _('Please correct the error below.')}
    pk_name = 'pk'

    permission = PermissionsSet(
        get=CoreReadAllowed() | CoreUpdateAllowed(),
        post=CoreUpdateAllowed(),
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    detail_verbose_name = None

    def get_detail_verbose_name(self):
        return self.model._ui_meta.detail_verbose_name if self.detail_verbose_name is None else self.detail_verbose_name

    def get_title(self):
        return (self.title or
                self.get_detail_verbose_name() % {
                    'verbose_name': self.model._meta.verbose_name,
                    'verbose_name_plural': self.model._meta.verbose_name_plural,
                    'obj': self.get_obj(True)
                })

    def is_readonly(self):
        return not self.has_permission('post')

    def link(self, arguments=None, **kwargs):
        if arguments is None:
            arguments = (self.kwargs[self.pk_name],)
        return super().link(arguments=arguments, **kwargs)

    # TODO: get_obj should not be inside core get_obj and _get_perm_obj_or_404 should have same implementation
    def _get_perm_obj_or_404(self, pk=None):
        """
        If is send parameter pk is returned object according this pk,
        else is returned object from get_obj method, but it search only inside filtered values for current user,
        finally if object is still None is returned according the input key from all objects.

        If object does not exist is raised Http404
        """
        if pk:
            obj = get_object_or_none(self.core.model, pk=pk)
        else:
            try:
                obj = self.get_obj(False)
            except Http404:
                obj = get_object_or_none(self.core.model, **self.get_obj_filters())
        if not obj:
            raise Http404
        return obj

    def _get_export_types(self):
        return self.core.get_export_types(self.request) if self.export_types is None else self.export_types

    def _get_export_fields(self):
        return list(self.core.get_export_fields(self.request, self.get_obj(True)))

    def _generate_rest_detail_export_fieldset(self):
        return ModelFlatRestFields.create_from_flat_list(self._get_export_fields(), self.model)

    def get_context_data(self, form=None, inline_form_views=None, **kwargs):
        context_data = super().get_context_data(
            form=form, inline_form_views=inline_form_views, **kwargs
        )
        if self._get_export_types() and self._get_export_fields():
            context_data.update({
                'export_types': get_export_types_with_content_type(self._get_export_types()),
                'rest_detail_export_fieldset': self._generate_rest_detail_export_fieldset(),
                'api_url': self.core.get_api_detail_url(self.request, self.get_obj(True))
            })
        return context_data


class ModelReadonlyDetailView(GetModelObjectCoreViewMixin, FieldPermissionViewMixin, GetMethodFieldMixin,
                              ListParentMixin, DefaultModelCoreViewMixin, TemplateView):

    template_name = 'is_core/generic_views/readonly_detail.html'

    permission = PermissionsSet(
        get=CoreReadAllowed(),
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    fields = None
    fieldsets = None
    field_labels = None
    view_type = 'readonly-detail'
    detail_verbose_name = None
    inline_views = None

    def get_inline_views(self):
        return self.inline_views

    def get_detail_verbose_name(self):
        return self.detail_verbose_name

    def get_prefix(self):
        return '-'.join((self.view_type, self.site_name, self.core.get_menu_group_pattern_name())).lower()

    def is_readonly(self):
        return True

    def get_fields(self):
        return (
            self.fields is not None and self.fields or
            self.core.get_fields(self.request, self.get_obj())
        )

    def get_fieldsets(self):
        return (
            self.fieldsets is not None and self.fieldsets or
            self.core.get_fieldsets(self.request)
        )

    def generate_fieldsets(self, obj):
        fieldsets = self.get_fieldsets()

        if fieldsets is None:
            fieldsets = [
                (None, {
                    'fields': self.get_fields() or []
                })
            ]
            for inline_view in self.get_inline_views() or ():
                inline_view_inst = (
                    inline_view(self.request, self, obj) if isinstance(inline_view, type) else inline_view
                )
                if inline_view_inst.can_render():
                    # Only inline view that can be rendered is added to formset
                    fieldsets.append((
                        inline_view_inst.get_title(), {
                            'inline_view': inline_view,
                            'inline_view_inst': inline_view_inst
                        }
                    ))
        else:
            inline_view_opts = get_inline_views_opts_from_fieldsets(fieldsets)
            for inline_view_opt in inline_view_opts:
                inline_view = inline_view_opt['inline_view']
                inline_view_inst = (
                    inline_view(self.request, self, obj) if isinstance(inline_view, type) else inline_view
                )
                if inline_view_inst.can_render():
                    # Only inline view that can be rendered is added to formset
                    inline_view_opt['inline_view_inst'] = inline_view_inst

        return get_fieldsets_without_disallowed_fields(
            self.request, fieldsets, self._get_disallowed_fields_from_permissions()
        )

    def _get_field_labels(self):
        return (
            self.field_labels if self.field_labels is not None else self.core.get_field_labels(self.request)
        )

    def _get_field_to_render(self, obj, field_name):
        value, label, widget = get_readonly_field_data(
            obj, field_name, self.request, view=self, field_labels=self._get_field_labels()
        )
        return field_name, label, display_for_value(value, request=self.request), None

    def _get_fieldset_to_render(self, fieldset_title, fieldset_data, obj):
        fields = [
            self._get_field_to_render(obj, field_name)
            for field_name in fieldset_data.get('fields', ())
        ]
        fieldsets = [
            self._get_fieldset_to_render(sub_fieldset_title, sub_fieldset_data, obj)
            for sub_fieldset_title, sub_fieldset_data in fieldset_data.get('fieldsets', ())
        ]

        inline_view = fieldset_data.get('inline_view_inst')
        rendered_inline_view = inline_view.render(RequestContext(self.request), fieldset_title) if inline_view else ''

        return {
            'title': fieldset_title,
            'fields': fields,
            'fieldsets': fieldsets,
            'class': fieldset_data.get('class'),
            'rendered_inline_view': rendered_inline_view
        }

    def _get_fieldsets_to_render(self):
        obj = self.get_obj()
        return [
            self._get_fieldset_to_render(fieldset_title, fieldset_data, obj)
            for fieldset_title, fieldset_data in self.generate_fieldsets(obj)
        ]

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        context_data.update({
            'fieldsets': self._get_fieldsets_to_render()
        })
        return context_data


class DjangoReadonlyDetailView(GetDjangoObjectCoreViewMixin, ModelReadonlyDetailView):

    def get_detail_verbose_name(self):
        return self.model._ui_meta.detail_verbose_name if self.detail_verbose_name is None else self.detail_verbose_name


class DjangoRelatedCoreTableView(DjangoReadonlyDetailView):

    table_model = None

    @cached_property
    def inline_table_view(self):
        return type(
            self.__class__.__name__ + 'InlineTableView', (DjangoInlineTableView,), {'model': self.table_model}
        )

    def get_fieldsets(self):
        return (
            (None, {'inline_view': self.inline_table_view}),
        )



from copy import deepcopy


class MenuItem:

    def __init__(self, title, collapsible, group, active=False, submenu_items=[]):
        self.title = title
        self.collapsible = collapsible
        self.submenu_items = submenu_items
        self.group = group
        self.active = active


class LinkMenuItem(MenuItem):

    def __init__(self, title, url, group=None, active=False, submenu_items=[]):
        super(LinkMenuItem, self).__init__(title, False, group, active, submenu_items)
        self.url = url

    def __str__(self):
        return self.title


class CollapsibleMenuItem(MenuItem):

    def __init__(self, title, url=None, group=None, active=False, submenu_items=[]):
        super(CollapsibleMenuItem, self).__init__(title, True, group, active, submenu_items)
        self.url = url


class MenuGenerator:

    default_menu_item_class = LinkMenuItem

    def __init__(self, request, site, active_groups):
        self.request = request
        self.site = site
        self.active_groups = active_groups

    def get_core(self, core_name):
        from .main import UiCore

        core = self.site._registry.get(core_name)
        return core if isinstance(core, UiCore) else None

    def get_menu_item(self, core_name, group=None):
        core = self.get_core(core_name)
        return core.get_menu_item(self.request, group, self.default_menu_item_class) if core else None

    def get_menu_items(self, items):
        if self.active_groups:
            group = self.active_groups[0]
        else:
            group = None

        for item in items:
            if isinstance(item, MenuItem):
                menu_item = deepcopy(item)
                if menu_item.submenu_items:
                    menu_item.submenu_items = list(self.get_menu_items(item.submenu_items))

                yield menu_item
            else:
                menu_item = self.get_menu_item(item, group)
                if menu_item:
                    yield menu_item

    def get_menu_structure(self):
        return self.site._registry.keys()



from django.shortcuts import resolve_url

from is_core.contrib.dynamo.cores import DynamoUiRestCore
from is_core.patterns import RestPattern, UiPattern, PK_PATTERN

from .models import Comment


class CommentPatternMixin:

    def _get_try_kwargs(self, request, obj):
        if obj and PK_PATTERN in self.url_pattern:
            return {
                'issue_pk': obj.issue_id,
                'pk': obj.user_id
            }
        else:
            return {'issue_pk': obj.issue_id if obj else 0}


class CommentRestPattern(CommentPatternMixin, RestPattern):
    pass


class CommentUiPattern(CommentPatternMixin, UiPattern):
    pass


class DynamoCommentCore(DynamoUiRestCore):

    model = Comment
    menu_group = 'dynamo-comment'

    fields = ('user_id', 'content', 'is_public', 'priority')

    verbose_name = 'comment'
    verbose_name_plural = 'comments'

    default_rest_pattern_class = CommentRestPattern
    default_ui_pattern_class = CommentUiPattern

    rest_range_key = 'user_id'

    def _get_hash_key(self, request):
        return request.kwargs.get('issue_pk')

    def get_api_url(self, request):
        return resolve_url(self.get_api_url_name(), issue_pk=self._get_hash_key(request))

    def get_url_prefix(self):
        return '/'.join(list(self.get_menu_groups()) + ['(?P<issue_pk>[^/]+)'])



from is_core.rest.resource import ModelCoreResourceMixin

from pyston.contrib.dynamo.resource import BaseDynamoResource


class DynamoCoreResource(ModelCoreResourceMixin, BaseDynamoResource):

    def get_range_key(self):
        return self.core.rest_range_key if self.range_key is None else self.range_key

    def _get_hash_key(self):
        return self.core._get_hash_key(self.request)

    def _get_obj_or_none(self, pk=None):
        return self.core.get_obj(self.request, pk)



# Generated by Django 3.1.7 on 2021-04-01 09:28

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('issue_tracker', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='related_object_ct',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AddField(
            model_name='issue',
            name='related_object_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



import django.db.models.options as options

from django.db import models
from django.db.models.fields import URLField
from django.db.models.fields.related import ForeignKey, ManyToManyField, OneToOneField
from django.utils.translation import ugettext_lazy as _

from chamber.patch import Options

from .humanize import url_humanized


options.DEFAULT_NAMES = options.DEFAULT_NAMES + ('default_fk_filter', 'default_m2m_filter', 'default_rel_filter')


class UIOptions(Options):

    model_class = models.Model
    meta_class_name = 'UIMeta'
    meta_name = '_ui_meta'
    attributes = {
        'extra_selecbox_fields': {},
        'list_verbose_name': '%(verbose_name_plural)s',
        'add_verbose_name': _('add %(verbose_name)s'),
        'add_button_verbose_name': _('add %(verbose_name)s'),
        'add_inline_button_verbose_name': _('add %(verbose_name)s'),
        'detail_verbose_name': '%(obj)s',
        'filter_placeholders': {},
        'placeholders': {},
        'default_ui_filter_by': None,
    }


def fk_formfield(self, **kwargs):
    from is_core import forms as is_forms

    kwargs.setdefault('form_class', is_forms.ModelChoiceField)
    return self._is_core_formfield_tmp(**kwargs)


def m2m_formfield(self, **kwargs):
    from is_core import forms as is_forms

    kwargs.setdefault('form_class', is_forms.ModelMultipleChoiceField,)
    return self._is_core_formfield_tmp(**kwargs)


def rel_field_init(self, *args, **kwargs):
    self._reverse_verbose_name = kwargs.pop('reverse_verbose_name', None)
    self._is_core_init_tmp(*args, **kwargs)


def get_reverse_verbose_name_plural(self):
    return self._reverse_verbose_name or self.model._meta.verbose_name_plural


def get_reverse_verbose_name(self):
    return self._reverse_verbose_name or self.model._meta.verbose_name


ForeignKey._is_core_formfield_tmp = ForeignKey.formfield
ForeignKey.formfield = fk_formfield
ForeignKey._is_core_init_tmp = ForeignKey.__init__
ForeignKey.__init__ = rel_field_init
ForeignKey.reverse_verbose_name = property(get_reverse_verbose_name_plural)

OneToOneField.reverse_verbose_name = property(get_reverse_verbose_name)

ManyToManyField._is_core_formfield_tmp = ManyToManyField.formfield
ManyToManyField.formfield = m2m_formfield
ManyToManyField._is_core_init_tmp = ManyToManyField.__init__
ManyToManyField.__init__ = rel_field_init
ManyToManyField.reverse_verbose_name = property(get_reverse_verbose_name_plural)

URLField.default_humanized = url_humanized
